/*
 * livepatch_bsc1198590
 *
 * Fix for CVE-2022-1280, bsc#1198590
 *
 *  Upstream commits:
 *  869e76f7a918 ("drm: avoid circular locks in drm_mode_getconnector")
 *  5eff9585de22 ("drm: avoid blocking in drm_clients_info's rcu section")
 *  1f7ef07cfa14 ("drm: add a locked version of drm_is_current_master")
 *  0b0860a3cf5e ("drm: serialize drm_file.master with a new spinlock")
 *  56f0729a510f ("drm: protect drm_master pointers in drm_lease.c")
 *  28be2405fb75 ("drm: use the lookup lock in drm_is_current_master")
 *
 *  SLE12-SP3 commit:
 *  Not affected
 *
 *  SLE12-SP4, SLE15 and SLE15-SP1 commit:
 *  Not affected
 *
 *  SLE12-SP5 commits:
 *  3095d7b2e6f0f67718f4aab263017161e9dcc474
 *  4f97fa9204c6a6298b014719f8f29a5ad9e6fee0
 *  16fea17c857feea5c62220fe571004ad946fab98
 *  4b9807ba7b0dbb9108b7df4bfa5ac24eb42e8ca9
 *
 *  SLE15-SP2 commits:
 *  cb6322a7d6cfd5a1b29353818ab06ff5dd563702
 *  a42efa862001c7a8051dd6a2d36dc1ccd81e4187
 *  cd36b1c933630df271fd5297cdb74c573bcd88d6
 *  f800e5380a3b19bd843d3e9166f5e2e4d999998f
 *  82a498ae7f940156157ba6f50140a8fe61a40635
 *
 *  SLE15-SP3 commits:
 *  cb6322a7d6cfd5a1b29353818ab06ff5dd563702
 *  195575e704a4442e30a3afe22b84a01da9e4fdcd
 *  81bf8e60f8221875ea950a6c05327ea0ab896ef5
 *  d3c944d8574354746dd8a23531722b0bf099cc91
 *  05fda16d52b2fcea21717f3d2364427eb58b1a50
 *
 *  SLE15-SP4 commits:
 *  b21a4c051d6873b128dc14588a6d8ad6dd31a88b
 *  2b83b5007716286e7365dc31f1588b5665824d5b
 *  0dfa2a0a001950371cc73e764afdf287aa51629c
 *  c15b99f135e9036c4ca520eb303c1c9ea7cf76c7
 *  459d36fb11e19207805d9bf649914eea50dc5e1a
 *  425221af7c35a82db00cebd88fec4e6ebba503e1
 *
 *
 *  Copyright (c) 2022 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if !IS_MODULE(CONFIG_DRM)
#error "Live patch supports only CONFIG_DRM=m"
#endif

/* klp-ccp: from drivers/gpu/drm/drm_auth.c */
#include <linux/slab.h>
#include <drm/drm_auth.h>

/* klp-ccp: from include/drm/drm_auth.h */
bool klpp_drm_is_current_master(struct drm_file *fpriv);

/* klp-ccp: from drivers/gpu/drm/drm_auth.c */
#include <drm/drm_drv.h>
#include <drm/drm_file.h>

/* klp-ccp: from include/drm/drm_lease.h */
static struct drm_master *(*klpe_drm_lease_owner)(struct drm_master *master);

/* klp-ccp: from drivers/gpu/drm/drm_legacy.h */
#include <linux/list.h>
#include <drm/drm.h>
#include <drm/drm_device.h>

/* klp-ccp: from drivers/gpu/drm/drm_auth.c */
bool klpp_drm_is_current_master(struct drm_file *fpriv)
{
	/*
	 * Fix CVE-2022-1280
	 *  -1 line, +7 lines
	 */
	bool ret;

	spin_lock(&fpriv->master_lookup_lock);
	ret = fpriv->is_master && (*klpe_drm_lease_owner)(fpriv->master) == fpriv->minor->dev->master;
	spin_unlock(&fpriv->master_lookup_lock);

	return ret;
}



#include <linux/kernel.h>
#include <linux/module.h>
#include "livepatch_bsc1198590.h"
#include "../kallsyms_relocs.h"

#define LIVEPATCHED_MODULE "drm"

static struct klp_kallsyms_reloc klp_funcs[] = {
	{ "drm_lease_owner", (void *)&klpe_drm_lease_owner, "drm" },
};

static int livepatch_bsc1198590_module_notify(struct notifier_block *nb,
					      unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, LIVEPATCHED_MODULE))
		return 0;

	ret = klp_resolve_kallsyms_relocs(klp_funcs, ARRAY_SIZE(klp_funcs));
	WARN(ret, "livepatch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block livepatch_bsc1198590_module_nb = {
	.notifier_call = livepatch_bsc1198590_module_notify,
	.priority = INT_MIN+1,
};

int livepatch_bsc1198590_init(void)
{
	int ret;
	struct module *mod;

	ret = klp_kallsyms_relocs_init();
	if (ret)
		return ret;

	ret = register_module_notifier(&livepatch_bsc1198590_module_nb);
	if (ret)
		return ret;

	rcu_read_lock_sched();
	mod = (*klpe_find_module)(LIVEPATCHED_MODULE);
	if (!try_module_get(mod))
		mod = NULL;
	rcu_read_unlock_sched();

	if (mod) {
		ret = klp_resolve_kallsyms_relocs(klp_funcs,
						  ARRAY_SIZE(klp_funcs));
	}

	if (ret)
		unregister_module_notifier(&livepatch_bsc1198590_module_nb);

	module_put(mod);
	return ret;
}

void livepatch_bsc1198590_cleanup(void)
{
	unregister_module_notifier(&livepatch_bsc1198590_module_nb);
}
